import { requireModule, resolveModule } from '@nuxt/kit';
import jiti from 'jiti';
import chroma from 'chroma-js';
import { existsSync } from 'fs';
import { mkdir, rm, writeFile } from 'fs/promises';
import StyleDictionary from 'browser-style-dictionary/browser.js';
import chalk from 'chalk';
import { kebabCase } from 'scule';
import consola from 'consola';
import { resolve } from 'pathe';
import { defu } from 'defu';

const name = "@nuxtjs/design-tokens-edge";
const version = "0.0.1-27696452.c073083";

const MODULE_DEFAULTS = {
  tokens: true,
  server: false
};
const _logger = consola;
function useLogger(scope) {
  return scope ? _logger.withScope(scope) : logger;
}
const logger = useLogger("design-tokens");
chalk.magentaBright(name);
const objectPaths = (data) => {
  const output = [];
  function step(obj, prev) {
    Object.keys(obj).forEach((key) => {
      const value = obj[key];
      const isarray = Array.isArray(value);
      const type = Object.prototype.toString.call(value);
      const isobject = type === "[object Object]" || type === "[object Array]";
      const newKey = prev ? `${prev}.${key}` : key;
      if (!output.includes(newKey)) {
        output.push(newKey);
      }
      if (!isarray && isobject && Object.keys(value).length) {
        return step(value, newKey);
      }
    });
  }
  step(data);
  return output;
};
const resolveVariableFromPath = (path) => `var(--${path.split(".").map((key) => kebabCase(key)).join("-")})`;
const get = (obj, path, defValue = void 0) => {
  if (!path) {
    return void 0;
  }
  const pathArray = Array.isArray(path) ? path : path.match(/([^[.\]])+/g);
  const result = pathArray.reduce(
    (prevObj, key) => prevObj && prevObj[key],
    obj
  );
  return result === void 0 ? defValue : result;
};

const referencesRegex = new RegExp(
  "\\{([^}]+)\\}",
  "g"
);
const walkTokens = (obj, typing = true, aliased = {}) => {
  let type = {};
  if (obj.value) {
    const _path = obj.path.join(".");
    const keyRegex = /{(.*)}/g;
    const testOriginal = obj.original.value.match(referencesRegex);
    if (testOriginal?.[0] && testOriginal[0] === obj.original.value) {
      obj.value = obj.original.value.replace(keyRegex, (_, tokenPath) => {
        aliased[_path] = resolveVariableFromPath(tokenPath);
        return aliased[_path];
      });
    }
    obj.variable = aliased?.[_path] ? aliased[_path] : `var(--${obj.name})`;
    type = typing ? "DesignToken" : obj;
  } else {
    for (const k in obj) {
      if (obj[k] && typeof obj[k] === "object") {
        type[k] = walkTokens(obj[k], typing, aliased).type;
      }
    }
  }
  return { type, aliased };
};

const tokensHelper = (ts = false) => {
  return `const defaultTokensHelperOptions${ts ? ": TokensHelperOptions" : ""} = {
  key: 'variable',
  flatten: true,
  silent: false
}

/**
 * Get a theme token by its path
 */
export const $tokens = (path${ts ? ": NuxtThemeTokensPaths" : ""} = undefined, options${ts ? ": TokensHelperOptions" : ""} = {}) => {
  const { key, flatten } = Object.assign(defaultTokensHelperOptions, options)

  if (!path) return themeTokens

  if (key === 'variable' && tokensAliases[path]) {
    return tokensAliases[path]
  }

  const token = get(themeTokens, path)

  if (key && token?.[key]) { return token[key] }

  if (key && flatten && typeof token === 'object') {
    const flattened = {}
    
    const flattenTokens = (obj) => {
      Object.entries(obj).forEach(([objectKey, value]) => {
        if (value[key]) {
          flattened[objectKey] = value[key]
          return
        }

        flattenTokens(value)
      })
    }

    flattenTokens(token)

    return flattened
  }

  return token
}

`;
};

const getFunction = `const get = ${get.toString()}`;
const DesignTokenType = `interface DesignToken<T = string | number> {
  /* The raw value you specified in your token declaration. */
  value?: T;
  /* CSS Variable reference that gets generated. */
  variable?: string;
  name?: string;
  comment?: string;
  themeable?: boolean;
  property?: string;
  /* Is the property using compose() */
  composed?: boolean;
  /* Is the property using palette() */
  palette?: boolean
  /* Is the property nested in a selector */
  nested?: string;
  /* Is the property nested in variant */
  variant?: string;
  attributes?: {
    category?: string;
    type?: string;
    item?: string;
    subitem?: string;
    state?: string;
    [key: string]: any;
  };
  [key: string]: any;
}`;
const NuxtDesignTokensType = `interface NuxtDesignTokens {
  [key: string]: NuxtDesignTokens | DesignToken;
}`;
const TokenHelperOptionsType = `interface TokenHelperOptions {
  /**
   * The key that will be unwrapped from the design token object.
   * @default variable
   */
  key?: string
  /**
   * Toggle logging if requesting an unknown token.
   * @default false
   */
  silent?: boolean
  /**
   * Toggle deep flattening of the design token object to the requested key.
   *
   * If you query an token path containing mutliple design tokens and want a flat \`key: value\` object, this option will help you do that.
   */
  flatten?: boolean
}`;

const tsTypesDeclaration = ({ tokens }) => {
  const { type } = walkTokens(tokens);
  let result = "import type { Ref\xA0} from 'vue'\n\n";
  result = "import { GlobalTokens } from '@nuxtjs/design-tokens'\n\n";
  result = result + `export ${DesignTokenType}

`;
  result = result + `export ${NuxtDesignTokensType}

`;
  result = result + `export ${TokenHelperOptionsType}

`;
  result = result + `export interface NuxtThemeTokens extends GlobalTokens ${JSON.stringify(type, null, 2)}

`;
  const tokensPaths = objectPaths(type);
  if (tokensPaths.length) {
    result = result + `export type NuxtThemeTokensPaths = ${tokensPaths.map((path) => `'${path}'`).join(" | \n")}

`;
  } else {
    result = result + "export type NuxtThemeTokensPaths = 'no.tokens'\n\n";
  }
  result = result.replace(/"DesignToken"/g, "DesignToken");
  return result;
};
const tsFull = ({ tokens }) => {
  const { type, aliased } = walkTokens(tokens, false);
  let result = "import type { NuxtThemeTokens,   NuxtThemeTokensPaths, TokenHelperOptions } from './types.d'\n\n";
  result = result + "export * from './types.d'\n\n";
  result = result + `${getFunction}

`;
  result = result + `export const tokensAliases = ${JSON.stringify(aliased, null, 2)} as const

`;
  result = result + `export const themeTokens: NuxtThemeTokens = ${JSON.stringify(type, null, 2)}

`;
  result = result + tokensHelper(true);
  result = result + "export const $dt = $tokens\n\n";
  return result;
};
const jsFull = ({ tokens }) => {
  const { type, aliased } = walkTokens(tokens, false);
  let result = `${getFunction}

`;
  result = result + `export const tokensAliases = ${JSON.stringify(aliased, null, 2)}

`;
  result = result + `export const themeTokens = ${JSON.stringify(type, null, 2)}

`;
  result = result + tokensHelper(false);
  result = result + "export const $dt = $tokens\n\n";
  return result;
};

const resolveConfig = (layer, key, configFile = `${key}.config`) => {
  const value = layer.config?.style?.[key] || MODULE_DEFAULTS[key];
  let config = {};
  let filePath;
  if (typeof value === "boolean") {
    filePath = resolve(layer.cwd, configFile);
  } else if (typeof value === "string") {
    filePath = resolve(layer.cwd, value);
  } else if (typeof value === "object") {
    config = value;
  }
  if (filePath) {
    try {
      const _file = requireModule(filePath, { clearCache: true });
      if (_file) {
        config = _file;
      }
    } catch (_) {
    }
  }
  return { filePath, config };
};
const resolveConfigTokens = (layers) => {
  const tokensFilePaths = [];
  let tokens = {};
  const splitLayer = (layer) => {
    if (layer.config?.style?.tokens || MODULE_DEFAULTS.tokens) {
      const { config: layerTokens, filePath: _layerTokensFilePath } = resolveConfig(layer, "tokens", "tokens.config");
      if (_layerTokensFilePath) {
        tokensFilePaths.push(_layerTokensFilePath);
      }
      tokens = defu(tokens, layerTokens);
    }
  };
  for (const layer of layers) {
    splitLayer(layer);
  }
  return { tokensFilePaths, tokens };
};
const createTokensDir = async (path) => {
  if (!existsSync(path)) {
    await mkdir(path, { recursive: true });
  }
};

const stubTokens = async (buildPath, force = false) => {
  const files = {
    "__uses-tokens.ts": () => "",
    "tokens.css": () => "/* This file is empty because no tokens has been provided. */",
    "tokens.scss": () => "/* This file is empty because no tokens has been provided. */",
    "tokens.json": () => "{}",
    "index.js": jsFull,
    "index.ts": tsFull,
    "types.d.ts": tsTypesDeclaration
  };
  for (const [file, stubbingFunction] of Object.entries(files)) {
    const path = `${buildPath}${file}`;
    if (force && existsSync(path)) {
      await rm(path);
    }
    if (!existsSync(path)) {
      await writeFile(path, stubbingFunction ? stubbingFunction({ tokens: {}, allTokens: [] }) : "");
    }
  }
};
const getStyleDictionaryInstance = async (tokens, buildPath) => {
  let styleDictionary = StyleDictionary;
  styleDictionary.fileHeader = {};
  styleDictionary.registerTransformGroup({
    name: "tokens-js",
    transforms: ["name/cti/kebab", "size/px", "color/hex"]
  });
  styleDictionary.registerFormat({
    name: "typescript/types-declaration",
    formatter({ dictionary }) {
      return tsTypesDeclaration(dictionary);
    }
  });
  styleDictionary.registerFormat({
    name: "css/full",
    formatter({ dictionary, options }) {
      const selector = options.selector ? options.selector : ":root";
      const { outputReferences } = options;
      const { formattedVariables } = StyleDictionary.formatHelpers;
      const { aliased } = walkTokens(dictionary.tokens, false);
      dictionary.allTokens = dictionary.allTokens.filter((token) => !aliased[token.path.join(".")]);
      return `${selector} {
` + formattedVariables({ format: "css", dictionary, outputReferences }) + "\n}\n";
    }
  });
  styleDictionary.registerFormat({
    name: "typescript/full",
    formatter({ dictionary }) {
      return tsFull(dictionary);
    }
  });
  styleDictionary.registerFormat({
    name: "javascript/full",
    formatter({ dictionary }) {
      return jsFull(dictionary);
    }
  });
  styleDictionary = await styleDictionary.extend({
    tokens,
    platforms: {
      scss: {
        transformGroup: "tokens-js",
        buildPath,
        files: [
          {
            destination: "tokens.scss",
            format: "scss/variables"
          }
        ]
      },
      json: {
        transformGroup: "tokens-js",
        buildPath,
        files: [
          {
            destination: "tokens.json",
            format: "json/flat"
          }
        ]
      },
      ts: {
        transformGroup: "tokens-js",
        buildPath,
        files: [
          {
            destination: "index.ts",
            format: "typescript/full"
          },
          {
            destination: "types.d.ts",
            format: "typescript/types-declaration"
          }
        ]
      },
      js: {
        transformGroup: "tokens-js",
        buildPath,
        files: [
          {
            destination: "index.js",
            format: "javascript/full"
          }
        ]
      },
      css: {
        transformGroup: "tokens-js",
        buildPath,
        files: [
          {
            destination: "tokens.css",
            format: "css/full"
          }
        ]
      },
      done: {
        actions: ["done"]
      }
    }
  });
  return styleDictionary;
};
const generateTokensOutputs = (styleDictionary, silent = true) => new Promise(
  (resolve, reject) => {
    try {
      if (silent) {
        console._log = console.log;
        console.log = () => {
        };
      }
      styleDictionary.registerAction({
        name: "done",
        do: (dictionary) => {
          resolve(dictionary);
        },
        undo: () => {
        }
      });
      styleDictionary.cleanAllPlatforms();
      styleDictionary.buildAllPlatforms();
      if (silent) {
        console.log = console._log;
      }
    } catch (e) {
      reject(e);
    }
  }
);
const generateTokens = async (tokens, buildPath, silent = true) => {
  if (!existsSync(buildPath)) {
    await createTokensDir(buildPath);
  }
  if (!tokens || !Object.keys(tokens).length) {
    await stubTokens(buildPath, true);
    return;
  }
  const styleDictionary = await getStyleDictionaryInstance(tokens, buildPath);
  return await generateTokensOutputs(styleDictionary, silent);
};

const palette = (color, suffixes = [50, 100, 200, 300, 400, 500, 600, 700, 800, 900], padding = 0.1) => {
  if (!color || typeof color !== "string") {
    throw new Error('Please provide a valid "color" string parameter');
  }
  function scalePalette(baseColor, _suffixes = suffixes, _padding = padding) {
    const colorscale = chroma.scale(["white", baseColor, "black"]).padding(padding).colors(suffixes.length);
    const colorRange = {};
    suffixes.forEach(
      (suffix, index) => colorRange[suffix] = {
        value: colorscale[index]
      }
    );
    colorRange[500] = {
      value: baseColor
    };
    return colorRange;
  }
  return {
    ...scalePalette(color),
    palette: true
  };
};

const defineTokens = (tokens) => tokens;

const css = (declaration) => declaration;

const $tokens = (path, options = {
  key: "variable",
  flatten: true,
  silent: false
}) => {
  const module = resolveModule(`${globalThis.__NuxtThemeTokensBuildDir__}index.ts`);
  const { $dt: $dt2 } = jiti(import.meta.url, { cache: false, requireCache: false, v8cache: false })(module);
  return $dt2(path, options);
};
const $dt = $tokens;

export { $tokens as $, MODULE_DEFAULTS as M, resolveConfigTokens as a, $dt as b, createTokensDir as c, defineTokens as d, getStyleDictionaryInstance as e, resolveConfig as f, generateTokens as g, css as h, logger as l, name as n, palette as p, referencesRegex as r, stubTokens as s, version as v };
