import { performance } from 'perf_hooks';
import { readFile } from 'fs/promises';
import { defineNuxtModule, createResolver, addPluginTemplate, resolveModule, addAutoImport } from '@nuxt/kit';
import { withTrailingSlash } from 'ufo';
import { defu } from 'defu';
import { join } from 'pathe';
import { $ as $tokens, l as logger, r as referencesRegex, n as name, v as version, M as MODULE_DEFAULTS, c as createTokensDir, g as generateTokens, a as resolveConfigTokens } from './chunks/index.mjs';
import fs from 'fs';
import { createUnplugin } from 'unplugin';
import { parse } from '@vue/compiler-sfc';
import MagicString from 'magic-string';
import { kebabCase, camelCase, upperFirst } from 'scule';
import chalk from 'chalk';
import * as recast from 'recast';
import json5 from 'json5';
import { stringify } from '@stitches/stringify';
import 'jiti';
import 'chroma-js';
import 'browser-style-dictionary/browser.js';
import 'consola';

function parseVueRequest(id) {
  const [filename, rawQuery] = id.split("?", 2);
  const params = new URLSearchParams(rawQuery);
  const ret = {};
  const langPart = Object.keys(Object.fromEntries(params)).find((key) => /lang\./i.test(key));
  ret.vue = params.has("vue") || id.endsWith(".vue");
  ret.global = params.has("global");
  ret.src = params.has("src");
  ret.raw = params.has("raw");
  if (params.has("type")) {
    ret.type = params.get("type");
  }
  if (params.has("blockType")) {
    ret.blockType = params.get("blockType");
  }
  if (params.has("index")) {
    ret.index = Number(params.get("index"));
  }
  if (params.has("data-v-transformed")) {
    ret.transformed = Boolean(params.get("data-v-transformed"));
  }
  if (params.has("scoped")) {
    ret.scoped = String(params.get("scoped"));
  }
  if (langPart) {
    const [, lang] = langPart.split(".");
    ret.lang = lang;
  } else if (params.has("lang")) {
    ret.lang = params.get("lang");
  }
  if (params.has("issuerPath")) {
    ret.issuerPath = params.get("issuerPath");
  }
  return {
    id,
    filename,
    query: ret
  };
}

const shortVariantsPropsRegex = /\$variantsProps\('(.*)'\)/g;
const fullVariantsPropsRegex = /\$variantsProps\(('(.*?)'),\s'(.*?)'\)/g;
const screenRegex = /(@screen\s(.*?)\s{)/g;
const darkRegex = /(@dark\s{)/g;
const lightRegex = /(@light\s{)/g;
const dtRegex = /\$dt\('(.*?)'\)/g;
const resolveStyle = (code = "") => {
  code = resolveDt(code);
  code = resolveScreens(code);
  code = resolveScheme(code, "dark");
  code = resolveScheme(code, "light");
  return code;
};
const resolveDt = (code, wrapper = void 0) => {
  const replace = (path) => `${wrapper || ""}var(--${path.split(".").map((key) => kebabCase(key)).join("-")})${wrapper || ""}`;
  code = code.replace(dtRegex, (_, path) => replace(path));
  return code;
};
const resolveVariantProps = (code = "", variantProps = {}) => {
  const propKeyToAst = (ast, key, prefix) => {
    ast.properties.push(
      recast.types.builders.objectProperty(
        recast.types.builders.stringLiteral(camelCase(camelCase(prefix + upperFirst(key)))),
        recast.types.builders.objectExpression([
          recast.types.builders.objectProperty(
            recast.types.builders.identifier("type"),
            recast.types.builders.identifier("Boolean")
          ),
          recast.types.builders.objectProperty(
            recast.types.builders.identifier("required"),
            recast.types.builders.identifier("false")
          ),
          recast.types.builders.objectProperty(
            recast.types.builders.identifier("default"),
            recast.types.builders.identifier("false")
          )
        ])
      )
    );
  };
  code = code.replace(
    shortVariantsPropsRegex,
    (_, tagName) => {
      const ast = recast.types.builders.objectExpression([]);
      if (variantProps[tagName]) {
        Object.keys(variantProps[tagName]).forEach(
          (propKey) => propKeyToAst(ast, propKey, "")
        );
      }
      return recast.print(ast).code;
    }
  );
  code = code.replace(
    fullVariantsPropsRegex,
    (_, tagName, prefix) => {
      const ast = recast.types.builders.objectExpression([]);
      if (variantProps[tagName]) {
        Object.keys(variantProps[tagName]).forEach(
          (propKey) => propKeyToAst(ast, propKey, prefix)
        );
      }
      return recast.print(ast).code;
    }
  );
  return code;
};
const resolveScheme = (code = "", scheme) => {
  const schemesRegex = {
    light: lightRegex,
    dark: darkRegex
  };
  code = code.replace(schemesRegex[scheme], `@media (prefers-color-scheme: ${scheme}) {`);
  return code;
};
const resolveScreens = (code = "") => {
  const screens = $tokens("screens", {
    flatten: false,
    key: void 0,
    silent: true
  });
  code = code.replace(
    screenRegex,
    (_, _screenDeclaration, screenSize) => {
      const screenToken = screens[screenSize];
      if (screenToken) {
        return `@media (min-width: ${screenToken.value}) {`;
      }
      logger.warn(
        "This screen size is not defined: " + chalk.red(screenSize) + "\n\n",
        "Available screen sizes: ",
        Object.keys(screens).map((screen) => chalk.green(screen)).join(", ")
      );
      return "@media (min-width: 0px) {";
    }
  );
  return code;
};

const cssContentRegex = /css\(({.*?\})\)/mgs;
const resolveStyleTs = (code = "", variantsProps = {}) => {
  const resolveVariantProps = (property, value) => {
    variantsProps[property] = Object.entries(value).reduce(
      (acc, [key, _]) => {
        acc[key] = {
          type: Boolean,
          required: false,
          default: false
        };
        return acc;
      },
      {}
    );
  };
  code = code.replace(
    cssContentRegex,
    (...cssFunctionMatch) => {
      const declaration = json5.parse(cssFunctionMatch[1]);
      const style = stringify(
        declaration,
        (property, value) => {
          if (property.startsWith("@")) {
            const DARK = "@dark";
            const LIGHT = "@light";
            const SCREEN = /@screen:(.*)/;
            const screenMatches = property.match(SCREEN);
            if (property === DARK) {
              return {
                "@media (prefers-color-scheme: dark)": value
              };
            }
            if (property === LIGHT) {
              return {
                "@media (prefers-color-scheme: light)": value
              };
            }
            if (screenMatches) {
              const screenToken = $tokens(`screens.${screenMatches[1]}`, { flatten: false, key: void 0, silent: true });
              return {
                [`@media (min-width: ${screenToken.original.value})`]: value
              };
            }
          }
          if (value.variants) {
            resolveVariantProps(property, value.variants);
          }
          if (property === "variants") {
            return Object.entries(value).reduce(
              (acc, [key, value2]) => {
                acc["&." + key] = value2;
                return acc;
              },
              {}
            );
          }
          if (typeof value === "string") {
            value = value.replace(
              referencesRegex,
              (...parts) => {
                const [, tokenPath] = parts;
                const token = $tokens(tokenPath);
                return token;
              }
            );
          }
          return {
            [property]: value
          };
        }
      );
      return style;
    }
  );
  return code;
};

const excludedPaths = [
  "node_modules/nuxt/dist",
  "node_modules/@vue/"
];
const unpluginNuxtStyle = createUnplugin(
  ({ components }) => {
    return {
      name: "unplugin-nuxt-design-tokens-style",
      enforce: "pre",
      load(id) {
        const { filename, query } = parseVueRequest(id);
        if (query.vue && query.type === "style") {
          const file = fs.readFileSync(filename, "utf8");
          const { descriptor } = parse(file, { filename });
          const style = descriptor.styles[query.index];
          let source = style.content || "";
          source = resolveStyleTs(source);
          source = resolveStyle(source);
          if (style.content !== source) {
            return source;
          }
        }
      },
      transformInclude(id) {
        const { query } = parseVueRequest(id);
        if (excludedPaths.some((path) => id.includes(path))) {
          return false;
        }
        if (components.some(({ filePath }) => id.includes(filePath))) {
          return true;
        }
        return query?.vue;
      },
      transform(code, id) {
        code = transformVueSFC(code, id);
        const { filename, query } = parseVueRequest(id);
        const magicString = new MagicString(code, { filename });
        const result = () => ({ code: magicString.toString(), map: magicString.generateMap() });
        try {
          if (query.type === "style") {
            code = resolveStyleTs(code);
            code = resolveStyle(code);
            return code;
          }
          const variantProps = {};
          const parsedComponent = parse(code, { filename });
          if (parsedComponent.descriptor.template) {
            const templateContent = parsedComponent.descriptor.template;
            let newTemplateContent = templateContent.content;
            newTemplateContent = resolveDt(newTemplateContent, "'");
            magicString.overwrite(templateContent.loc.start.offset, templateContent.loc.end.offset, newTemplateContent);
          }
          if (parsedComponent.descriptor.styles) {
            const styles = parsedComponent.descriptor.styles;
            styles.forEach(
              (styleBlock) => {
                const { loc, content } = styleBlock;
                let newStyle = content;
                newStyle = resolveStyleTs(newStyle, variantProps);
                newStyle = resolveStyle(newStyle);
                magicString.remove(loc.start.offset, loc.end.offset);
                magicString.appendRight(
                  loc.end.offset,
                  "\n" + newStyle + "\n"
                );
              }
            );
          }
          if (parsedComponent.descriptor.scriptSetup) {
            const scriptSetup = parsedComponent.descriptor.scriptSetup;
            let newScriptSetup = scriptSetup.content;
            newScriptSetup = resolveDt(newScriptSetup, "`");
            newScriptSetup = resolveVariantProps(newScriptSetup, variantProps);
            magicString.overwrite(scriptSetup.loc.start.offset, scriptSetup.loc.end.offset, newScriptSetup);
          }
        } catch (e) {
          console.log({ error: e });
        }
        return result();
      },
      vite: {
        handleHotUpdate(ctx) {
          const defaultRead = ctx.read;
          ctx.read = async function() {
            const code = await defaultRead();
            return transformVueSFC(code, ctx.file) || code;
          };
        }
      }
    };
  }
);
const registerTransformPlugin = (nuxt, options) => {
  nuxt.hook("webpack:config", (config) => {
    config.plugins = config.plugins || [];
    config.plugins.unshift(
      unpluginNuxtStyle.webpack(options)
    );
  });
  nuxt.hook("vite:extend", (vite) => {
    vite.config.plugins = vite.config.plugins || [];
    vite.config.plugins.push(
      unpluginNuxtStyle.vite(options)
    );
  });
};
function transformVueSFC(code, id) {
  if (id.endsWith(".vue") && !id.includes("?")) {
    const styleTagRe = /<style\b(.*?)\blang=['"][tj]sx?['"](.*?)>/;
    if (code.match(styleTagRe)) {
      return code.replace(styleTagRe, '<style$1lang="postcss"$2>');
    }
  }
  return code;
}

const module = defineNuxtModule({
  meta: {
    name,
    version,
    configKey: "style",
    compatibility: {
      nuxt: "^3.0.0-rc.5"
    }
  },
  defaults: MODULE_DEFAULTS,
  async setup(moduleOptions, nuxt) {
    nuxt.options.runtimeConfig.style = {
      tokensFilePaths: [],
      server: moduleOptions.server
    };
    let hasTailwind = false;
    nuxt.hook("tailwindcss:config", () => {
      hasTailwind = true;
    });
    let lastTokensCount = 0;
    const layers = nuxt.options._layers;
    const privateConfig = nuxt.options.runtimeConfig.style;
    const tokensDir = withTrailingSlash(nuxt.options.buildDir + "/tokens");
    globalThis.__NuxtThemeTokensBuildDir__ = tokensDir;
    const { resolve: resolveTokensDir } = createResolver(tokensDir);
    privateConfig.tokensDir = tokensDir;
    const { resolve: resolveRuntime } = createResolver(import.meta.url);
    const resolveRuntimeModule = (path) => resolveModule(path, { paths: resolveRuntime("./runtime") });
    await createTokensDir(tokensDir);
    const postcssOptions = nuxt.options.postcss || nuxt.options.build.postcss.postcssOptions || nuxt.options.build.postcss;
    postcssOptions.plugins = postcssOptions.plugins || {};
    postcssOptions.plugins["postcss-nested"] = postcssOptions.plugins["postcss-nested"] ?? {};
    postcssOptions.plugins["postcss-custom-properties"] = postcssOptions.plugins["postcss-custom-properties"] ?? {};
    const refreshTokens = async (nitro) => {
      const { tokensFilePaths: tokensFilePaths2, tokens: tokens2 } = resolveConfigTokens(layers);
      if (moduleOptions.tokens) {
        privateConfig.tokensFilePaths = tokensFilePaths2;
      }
      if (nitro) {
        const projectConfigPath = resolveModule(tokensFilePaths2[0]);
        const projectConfig = (await readFile(projectConfigPath)).toString();
        await nitro.storage.setItem("cache:design-tokens:config", projectConfig);
        await nitro.storage.setItem("cache:design-tokens:tokens.json", tokens2);
      }
      return { tokens: tokens2, tokensFilePaths: tokensFilePaths2 };
    };
    const buildTokens = async (nitro, viteServer) => {
      try {
        const start = performance.now();
        const tokens2 = await nitro.storage.getItem("cache:design-tokens:tokens.json");
        const dictionary = await generateTokens(tokens2, tokensDir);
        const end = performance.now();
        const ms = Math.round(end - start);
        logger.success(`Design Tokens built${ms ? " in " + ms + "ms" : ""}`);
        if (viteServer && lastTokensCount !== dictionary.allTokens.length) {
          lastTokensCount = dictionary.allTokens.length;
          viteServer.moduleGraph.invalidateAll();
          viteServer.ws.send({
            type: "full-reload",
            path: "*"
          });
        }
      } catch (e) {
        logger.error("Could not build tokens!");
        logger.error(e.message);
      }
    };
    const { tokens, tokensFilePaths } = await refreshTokens();
    nuxt.options.build.transpile = nuxt.options.build.transpile || [];
    nuxt.options.build.transpile.push(resolveRuntime("./runtime"));
    nuxt.options.alias = nuxt.options.alias || {};
    nuxt.options.alias["#design-tokens"] = resolveTokensDir("index");
    nuxt.options.alias["#design-tokens/style"] = resolveTokensDir("tokens.css");
    nuxt.options.alias["#design-tokens/types"] = resolveTokensDir("types.d");
    addPluginTemplate({
      filename: "tokens.mjs",
      getContents: () => {
        const lines = [
          `import '${resolveTokensDir("tokens.css")}'`,
          `import '${resolveRuntime("./runtime/assets/reset.css")}'`,
          "export default defineNuxtPlugin(() => {})"
        ];
        return lines.join("\n");
      }
    });
    nuxt.hook(
      "components:extend",
      (components) => {
        registerTransformPlugin(
          nuxt,
          {
            components: components.map(({ global, kebabName, pascalName, filePath }) => ({ global, kebabName, pascalName, filePath })),
            tokensDir,
            tokensFilePaths: tokensFilePaths.map((path) => resolveModule(path)),
            hasTailwind
          }
        );
      }
    );
    nuxt.hook("prepare:types", (opts) => {
      opts.references.push({ path: resolveTokensDir("types.d.ts") });
    });
    nuxt.hook("nitro:build:before", async (nitro) => {
      await refreshTokens(nitro);
      await buildTokens(nitro);
    });
    nuxt.hook("nitro:config", (nitroConfig) => {
      nitroConfig.handlers = nitroConfig.handlers || [];
      nitroConfig.handlers.push({
        method: "all",
        route: "/api/_design-tokens/tokens",
        handler: resolveRuntimeModule("./server/api/tokens/index")
      });
      nitroConfig.handlers.push({
        method: "all",
        route: "/api/_design-tokens/config",
        handler: resolveRuntimeModule("./server/api/tokens/config")
      });
      nitroConfig.prerender = nitroConfig.prerender || {};
      nitroConfig.prerender.routes = nitroConfig.prerender.routes || [];
      nitroConfig.prerender.routes.push("/api/_design-tokens/tokens");
      nitroConfig.prerender.routes.push("/api/_design-tokens/config");
      nitroConfig.bundledStorage = nitroConfig.bundledStorage || [];
      nitroConfig.bundledStorage.push("/cache/tokens");
      nitroConfig.bundledStorage.push("/tokens");
      if (privateConfig.server) {
        nitroConfig.handlers.push({
          method: "get",
          route: "/api/_design-tokens/tokens/generate",
          handler: resolveRuntimeModule("./server/api/tokens/generate")
        });
        nitroConfig.externals = defu(typeof nitroConfig.externals === "object" ? nitroConfig.externals : {}, {
          inline: [
            "lodash",
            "browser-style-dictionary",
            "browser-style-dictionary/lib",
            "browser-style-dictionary/browser",
            resolveRuntime()
          ]
        });
        nitroConfig.alias["#design-tokens"] = resolveRuntime("./runtime/server/utils");
      }
    });
    addAutoImport({
      from: resolveRuntimeModule("./composables/tokens"),
      name: "useTokens",
      as: "useTokens"
    });
    if (nuxt.options.dev) {
      nuxt.hook("nitro:init", (nitro) => {
        nuxt.hook("vite:serverCreated", (viteServer) => {
          tokensFilePaths.forEach((path) => viteServer.config.assetsInclude(resolveModule(path)));
          nuxt.hook("builder:watch", async (_, path) => {
            const extensions = [".js", ".ts", ".mjs"];
            const isTokensConfig = extensions.some((extension) => path.includes("tokens.config" + extension));
            if (isTokensConfig) {
              const { tokens: tokens2 } = await refreshTokens(nitro);
              viteServer.ws.send({
                type: "custom",
                event: "nuxt-design-tokens:update",
                data: {
                  tokens: moduleOptions.tokens ? tokens2 : {}
                }
              });
              await buildTokens(nitro, viteServer);
            }
          });
        });
      });
    }
    nuxt.hook("build:before", async () => await generateTokens(tokens, tokensDir, true));
    await generateTokens(tokens, tokensDir, true);
    nuxt.hook("tailwindcss:config", (tailwindConfig) => {
      tailwindConfig.content = tailwindConfig.content ?? [];
      tailwindConfig.content.push(join(tokensDir, "index.ts"));
    });
  }
});

export { module as default };
